#include <rt_config.h>

// Radar detection and carrier detection for 2880-SW-MCU

#if defined(DFS_SUPPORT) || defined(CARRIER_DETECTION_SUPPORT)

static inline void TimerCB_Radar(PRTMP_ADAPTER pAd);
static inline void TimerCB_Carrier(PRTMP_ADAPTER pAd);


void TimerCB(PRTMP_ADAPTER pAd)
{
	// what is this mean???
	if (pAd->CommonCfg.McuRadarProtection == 1)
	{
#ifdef DFS_DEBUG
		printk("###\n");
#endif // DFS_DEBUG //
		return;
	}

#ifdef CONFIG_AP_SUPPORT	
#ifdef DFS_SUPPORT
	if (pAd->CommonCfg.McuRadarCmd & RADAR_DETECTION)
		TimerCB_Radar(pAd);
#endif
#endif // CONFIG_AP_SUPPORT //

#ifdef CARRIER_DETECTION_SUPPORT
	if (pAd->CommonCfg.McuRadarCmd & CARRIER_DETECTION)
		TimerCB_Carrier(pAd);
#endif

}

#ifdef DFS_DEBUG
static unsigned int dbg_idx=0;

void DFS_Debug(UCHAR i, PRTMP_ADAPTER pAd)
{
	if (dbg_idx < 4*40)
	{
		pAd->CommonCfg.Detect1msIndex[dbg_idx] = i;
		pAd->CommonCfg.Detect1msType[dbg_idx] = pAd->CommonCfg.McuRadarTick;
		pAd->CommonCfg.Detect1msDuration[dbg_idx] = pAd->CommonCfg.McuCarrierTick;
		dbg_idx++;
	}
	if (dbg_idx == 4*40)
	{
		printk("data collect done\n");
		dbg_idx++;
	}
}

void DFS_Dump(PRTMP_ADAPTER pAd)
{
	int i;
	for (i = 0; i < 4 * 40; i++)
		printk("%x %d %d \n", pAd->CommonCfg.Detect1msIndex[i], pAd->CommonCfg.Detect1msType[i], pAd->CommonCfg.Detect1msDuration[i]);
	
	dbg_idx=0;
}
#endif // DFS_DEBUG //

#define CTS_WAIT_LOOP		500

#ifdef CONFIG_AP_SUPPORT
#ifdef DFS_SUPPORT
static inline void TimerCB_Radar(PRTMP_ADAPTER pAd)
{
	ULONG Value;

	if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CARRIER_DETECTING)
	{
		pAd->CommonCfg.McuRadarTick++;
		return;
	}
	else
		pAd->CommonCfg.McuRadarTick++;
	

	if (pAd->CommonCfg.McuRadarTick >= pAd->CommonCfg.McuRadarPeriod)
	{
		// Start of period
		pAd->CommonCfg.McuRadarTick = 0;
		pAd->CommonCfg.McuRadarDetectCount = 0;
		pAd->CommonCfg.McuRadarState = WAIT_CTS_BEING_SENT;
		pAd->CommonCfg.McuRadarEvent |= RADAR_EVENT_RADAR_DETECTING;


		if ((pAd->CommonCfg.McuRadarCtsProtect == 1)
#ifdef CARRIER_DETECTION_SUPPORT
			|| (isCarrierDetectExist(pAd))
#endif
			)
		{
			// no CTS protect
			pAd->CommonCfg.McuRadarState = DO_DETECTION;
			MCURadarDetect(pAd);

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x3, pAd);
			}
#endif // DFS_DEBUG //
			
		}
		else if (pAd->CommonCfg.McuRadarCtsProtect == 2 || pAd->CommonCfg.McuRadarCtsProtect == 3)
		{
			int i;
						
			if (pAd->CommonCfg.McuRadarCtsProtect == 3)
			{
				
				RTMP_IO_READ32(pAd, 0x7784, &Value);
				Value &= 0xffff00ff;
				RTMP_IO_WRITE32(pAd, 0x7784, Value);
				
				// pAd->CommonCfg.McuRadarCtsProtect == 3, need to kick CTS two times
				RTMP_IO_WRITE32(pAd, PBF_CTRL, 0x40);
				for (i = 0; i < CTS_WAIT_LOOP; i++)
				{
					RTMP_IO_READ32(pAd, PBF_CTRL, &Value);
					if (!Value)
						break;
				}
				
				if (i >= CTS_WAIT_LOOP)
				{
					pAd->CommonCfg.McuRadarEvent |= RADAR_EVENT_CTS_KICKED;
					return;
				}
			}


			RTMP_IO_READ32(pAd, 0x7784, &Value);
			Value &= 0xffff00ff;
			Value |= (DFS_CTS_WCID << 8);
			RTMP_IO_WRITE32(pAd, 0x7784, Value);

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x1, pAd);
			}
#endif // DFS_DEBUG //
			// kick CTS
			RTMP_IO_WRITE32(pAd, PBF_CTRL, 0x40);

		}
		return;
	}

	if (pAd->CommonCfg.McuRadarState == WAIT_CTS_BEING_SENT)
	{
		// check event or timeout
		if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CTS_SENT)
		{
			pAd->CommonCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_SENT;
			pAd->CommonCfg.McuRadarDetectCount = 1;
			pAd->CommonCfg.McuRadarState = DO_DETECTION;

			{
				UCHAR BBPR115;
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R115, &BBPR115);
			}

#ifdef DFS_DEBUG			
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x3, pAd);
			}
#endif // DFS_DEBUG //
			return;
		}
		
		if (pAd->CommonCfg.McuRadarDetectCount >= (5-1)) // timeout 5ms, start from 0
		{
			pAd->CommonCfg.McuRadarState = FREE_FOR_TX; // timeout 5ms, give up
			pAd->CommonCfg.McuRadarEvent &= ~(RADAR_EVENT_RADAR_DETECTING | RADAR_EVENT_CTS_KICKED);

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x5, pAd);
			}
#endif // DFS_DEBUG //

			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, pAd->CommonCfg.OldRtsRetryLimit);
			// Restore BBP
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->CommonCfg._R65);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);

			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
			
		}

		if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CTS_KICKED)
		{
			int i;

			for (i = 0; i < CTS_WAIT_LOOP; i++)
			{
				RTMP_IO_READ32(pAd, PBF_CTRL, &Value);
				if (!Value)
					break;
			}
				
			if (i >= CTS_WAIT_LOOP)
			{
				pAd->CommonCfg.McuRadarDetectCount++;
				return;
			}
			
			RTMP_IO_READ32(pAd, 0x7784, &Value);
			Value &= 0xffff00ff;
			Value |= (DFS_CTS_WCID << 8) ;
			RTMP_IO_WRITE32(pAd, 0x7784, Value);

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x1, pAd);
			}
#endif // DFS_DEBUG //

			// kick CTS
			RTMP_IO_WRITE32(pAd, PBF_CTRL, 0x40);
			
			pAd->CommonCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_KICKED;
		}

		
		pAd->CommonCfg.McuRadarDetectCount++;
	}
	else if (pAd->CommonCfg.McuRadarState == DO_DETECTION)
	{
		UCHAR BBPR115;
		
		BBP_IO_READ8_BY_REG_ID(pAd, BBP_R115, &BBPR115);

#ifdef DFS_FCC_BW40_FIX
		if ((pAd->CommonCfg.RadarDetect.RDDurRegion == FCC) && ((pAd->CommonCfg.McuRadarDetectCount == 2) || (pAd->CommonCfg.McuRadarDetectCount == 4) || (pAd->CommonCfg.McuRadarDetectCount == 7) || (pAd->CommonCfg.McuRadarDetectCount == 9)))
#else
		if ((pAd->CommonCfg.RadarDetect.RDDurRegion == FCC) && ((pAd->CommonCfg.McuRadarDetectCount % 2) == 0))
#endif
		{
			UCHAR bbp;
			
			if (BBPR115 == 0x2)
			{
				Value = BBPR115 << 24;
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R116, &bbp);
				Value |= (bbp << 16);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R117, &bbp);
				Value |= (bbp << 8);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R118, &bbp);
				Value |= bbp;
				RTMP_IO_WRITE32(pAd, 0x7100 + (4 * (pAd->CommonCfg.McuRadarDetectCount / 2 - 1)) , Value);
			
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x21);
			
				pAd->CommonCfg.McuRadarEvent |= RADAR_EVENT_WIDTH_RADAR;
			}
			else
			{
				RTMP_IO_WRITE32(pAd, 0x7100 + (4 * (pAd->CommonCfg.McuRadarDetectCount / 2 - 1)) , 0);
			}

#ifdef DFS_FCC_BW40_FIX
			if (pAd->CommonCfg.HtCapability.HtCapInfo.ChannelWidth  == BW_40)
			{
				if (pAd->CommonCfg.McuRadarDetectCount == ((pAd->CommonCfg.McuRadarDetectPeriod / 2) - 1))
				{
				   	ULONG		Value, byteValue = 0;

					// Switch to Extend Channel
					if (pAd->CommonCfg.AddHTInfo.AddHtInfo.ExtChanOffset == EXTCHA_BELOW)
					{
						//  RX : control channel move to lower 
						RTMP_BBP_IO_READ8_BY_REG_ID(pAd, 3, &byteValue);
						byteValue &= (~0x20);
						RTMP_BBP_IO_WRITE8_BY_REG_ID(pAd, 3, byteValue);
					}
					else
					{
						//  RX : control channel move to upper 
						RTMP_BBP_IO_READ8_BY_REG_ID(pAd, 3, &byteValue);
						byteValue |= (0x20);
						RTMP_BBP_IO_WRITE8_BY_REG_ID(pAd, 3, byteValue);
					}

					//DisableNormalTx
					RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
					RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, 0); // Toggle Tx Rx
					RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, 0xc); // Toggle Tx Rx
					RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);

				}
				
			}
#endif // DFS_FCC_BW40_FIX
		}
		else if ((pAd->CommonCfg.RadarDetect.RDDurRegion == JAP_W56) && ((pAd->CommonCfg.McuRadarDetectCount % 2) == 0))
		{
			UCHAR bbp;
			
			if (BBPR115 == 0x2)
			{
				Value = BBPR115 << 24;
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R116, &bbp);
				Value |= (bbp << 16);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R117, &bbp);
				Value |= (bbp << 8);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R118, &bbp);
				Value |= bbp;
				RTMP_IO_WRITE32(pAd, 0x7100 + (4 * (pAd->CommonCfg.McuRadarDetectCount / 2 - 1)) , Value);
			
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x21);
			
				pAd->CommonCfg.McuRadarEvent |= RADAR_EVENT_WIDTH_RADAR;
			}
			else
			{
				RTMP_IO_WRITE32(pAd, 0x7100 + (4 * (pAd->CommonCfg.McuRadarDetectCount / 2 - 1)) , 0);
			}
						
		}
		
		if (pAd->CommonCfg.McuRadarDetectCount >= (pAd->CommonCfg.McuRadarDetectPeriod - 1))
		{

#ifdef DFS_FCC_BW40_FIX
			// Go Back to Primary Channel
			if (pAd->CommonCfg.HtCapability.HtCapInfo.ChannelWidth  == BW_40)
			{
				ULONG		Value, byteValue = 0;

				if (pAd->CommonCfg.AddHTInfo.AddHtInfo.ExtChanOffset == EXTCHA_BELOW)
				{
					//  RX : control channel move to upper 
					RTMP_BBP_IO_READ8_BY_REG_ID(pAd, 3, &byteValue);
					byteValue |= (0x20);
					RTMP_BBP_IO_WRITE8_BY_REG_ID(pAd, 3, byteValue);
				}
				else
				{
					//  RX : control channel move to lower 
					RTMP_BBP_IO_READ8_BY_REG_ID(pAd, 3, &byteValue);
					byteValue &= (~0x20);
					RTMP_BBP_IO_WRITE8_BY_REG_ID(pAd, 3, byteValue);
				}
			}		
#endif // DFS_FCC_BW40_FIX

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x4, pAd);
			}
#endif // DFS_DEBUG //

			if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_WIDTH_RADAR)
			{
				pAd->CommonCfg.McuRadarEvent &= ~(RADAR_EVENT_WIDTH_RADAR);
				RadarSMDetect(pAd, RADAR_WIDTH);
			}
			
#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_SIMULATE) 
			{
				if (pAd->CommonCfg.McuRadarCmd & RADAR_DETECTION)
				{
					RadarSMDetect(pAd, RADAR_PULSE);
				}
			}
#endif // DFS_DEBUG //
			

			if (BBPR115 == 0x1)
			{
				
				// pulse radar detected
				if (pAd->CommonCfg.McuRadarCmd & RADAR_DETECTION)
				{
					RadarSMDetect(pAd, RADAR_PULSE);
				}
			}

			pAd->CommonCfg.McuRadarState = FREE_FOR_TX;
			pAd->CommonCfg.McuRadarEvent &= ~(RADAR_EVENT_RADAR_DETECTING);

			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, pAd->CommonCfg.OldRtsRetryLimit);
			// Restore BBP
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->CommonCfg._R65);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);

#ifdef DFS_DEBUG
			if ((pAd->CommonCfg.McuRadarDebug & RADAR_GPIO_DEBUG))
				_RTMP_IO_WRITE32(0xa0300630, 0x2000);
#endif // DFS_DEBUG //
			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);

		}
		pAd->CommonCfg.McuRadarDetectCount++;
	}
	else // FREE_FOR_TX
	{
		if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CTS_SENT)
		{

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x6, pAd);
			}
#endif // DFS_DEBUG //

			pAd->CommonCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_SENT;
			pAd->CommonCfg.McuRadarDetectCount = 1;

			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, pAd->CommonCfg.OldRtsRetryLimit);
			// Restore BBP
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->CommonCfg._R65);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);


#ifdef DFS_DEBUG
			if ((pAd->CommonCfg.McuRadarDebug & RADAR_GPIO_DEBUG))
				_RTMP_IO_WRITE32(0xa0300630, 0x2000);
#endif // DFS_DEBUG //
			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
		}
	}	
}
#endif // DFS_SUPPORT //
#endif // CONFIG_AP_SUPPORT //

static inline void TimerCB_Carrier(PRTMP_ADAPTER pAd)
{
	ULONG Value;

	if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_RADAR_DETECTING)
	{
		pAd->CommonCfg.McuCarrierTick++;
		return;
	}
	else
		pAd->CommonCfg.McuCarrierTick++;


	if (pAd->CommonCfg.McuCarrierTick >= pAd->CommonCfg.McuCarrierPeriod)
	{
		// Start of period
		pAd->CommonCfg.McuCarrierTick = 0;
		pAd->CommonCfg.McuCarrierDetectCount = 0;
		pAd->CommonCfg.McuCarrierState = WAIT_CTS_BEING_SENT;
		pAd->CommonCfg.McuRadarEvent |= RADAR_EVENT_CARRIER_DETECTING;

		if ((pAd->CommonCfg.McuCarrierCtsProtect == 1)
#ifdef CARRIER_DETECTION_SUPPORT
			|| (isCarrierDetectExist(pAd))
#endif // CARRIER_DETECTION_SUPPORT //
			)
		{
			// no CTS protect
			pAd->CommonCfg.McuCarrierState = DO_DETECTION;
			MCURadarDetect(pAd);

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x13, pAd);
			}
#endif // DFS_DEBUG //

		}
		else if (pAd->CommonCfg.McuCarrierCtsProtect == 2)
		{
#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x11, pAd);
			}
#endif // DFS_DEBUG //
			// kick CTS
			RTMP_IO_WRITE32(pAd, PBF_CTRL, 0x80);
		}
		return;
	}

	if (pAd->CommonCfg.McuCarrierState == WAIT_CTS_BEING_SENT)
	{
		// check event or timeout
		if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CTS_CARRIER_SENT)
		{
			pAd->CommonCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_CARRIER_SENT;
			pAd->CommonCfg.McuCarrierDetectCount = 1;
			pAd->CommonCfg.McuCarrierState = DO_DETECTION;

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x13, pAd);
			}
#endif // DFS_DEBUG //
			return;
		}
		
		if (pAd->CommonCfg.McuCarrierDetectCount >= (5-1)) // timeout 5ms, start from 0
		{
			pAd->CommonCfg.McuCarrierState = FREE_FOR_TX; // timeout 5ms, give up
			pAd->CommonCfg.McuRadarEvent &= ~(RADAR_EVENT_CARRIER_DETECTING);

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x15, pAd);
			}
#endif // DFS_DEBUG //
		}
		
		pAd->CommonCfg.McuCarrierDetectCount++;
	}
	else if (pAd->CommonCfg.McuCarrierState == DO_DETECTION)
	{
		if (pAd->CommonCfg.McuCarrierDetectCount >= (pAd->CommonCfg.McuCarrierDetectPeriod - 1))
		{
			UCHAR BBPR115;

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x14, pAd);
			}
#endif // DFS_DEBUG //

			BBP_IO_READ8_BY_REG_ID(pAd, BBP_R115, &BBPR115);

			CarrierDetectionCheck(pAd);

			pAd->CommonCfg.McuCarrierState = FREE_FOR_TX;
			pAd->CommonCfg.McuRadarEvent &= ~(RADAR_EVENT_CARRIER_DETECTING);

			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, pAd->CommonCfg.OldRtsRetryLimit);
			// Restore BBP

			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R69, pAd->CommonCfg._R69);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R70, pAd->CommonCfg._R70);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R73, pAd->CommonCfg._R73);

#ifdef DFS_DEBUG
			if ((pAd->CommonCfg.McuRadarDebug & RADAR_GPIO_DEBUG))
				_RTMP_IO_WRITE32(0xa0300630, 0x2000);
#endif // DFS_DEBUG //
			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);

			

		}
		pAd->CommonCfg.McuCarrierDetectCount++;
	}
	else // FREE_FOR_TX
	{
		if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CTS_CARRIER_SENT)
		{
#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(0x16, pAd);
			}
#endif // DFS_DEBUG // 
			pAd->CommonCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_CARRIER_SENT;
			pAd->CommonCfg.McuCarrierDetectCount = 1;

			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, pAd->CommonCfg.OldRtsRetryLimit);
			// Restore BBP
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R69, pAd->CommonCfg._R69);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R70, pAd->CommonCfg._R70);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R73, pAd->CommonCfg._R73);


#ifdef DFS_DEBUG
			if ((pAd->CommonCfg.McuRadarDebug & RADAR_GPIO_DEBUG))
				_RTMP_IO_WRITE32(0xa0300630, 0x2000);
#endif // DFS_DEBUG //
			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
		}
	}	
}


void inline MCURadarDetect(PRTMP_ADAPTER pAd)
{
	ULONG Value;
	UCHAR bbp;
	
			// Roger test
			BBP_IO_READ8_BY_REG_ID(pAd, BBP_R113, &bbp);
			if (bbp & 0x20)
				return;

			//DisableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, 0);
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, 0xc);
			Value &= (ULONG)~(0x04);
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
						
#ifdef DFS_DEBUG
			if ((pAd->CommonCfg.McuRadarDebug & RADAR_GPIO_DEBUG))
				if (pAd->CommonCfg.McuRadarCtsProtect != 1)
					_RTMP_IO_WRITE32(0xa030062c, 0x2000);
#endif // DFS_DEBUG //


			RTMP_IO_READ32(pAd, TX_RTS_CFG, &pAd->CommonCfg.OldRtsRetryLimit);
			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, 0);
			
			//change BBP now
			if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CARRIER_DETECTING)
			{
				UCHAR BBPR4;
				
				// Clear False CCA
				RTMP_IO_READ32(pAd, RX_STA_CNT1, &Value);
				
				// save old value
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R66, &pAd->CommonCfg._R66);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R69, &pAd->CommonCfg._R69);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R70, &pAd->CommonCfg._R70);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R73, &pAd->CommonCfg._R73);

				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R4, &BBPR4);
				if (BBPR4 & 0x18) // BW 40
				{
					if (pAd->CommonCfg.Channel > 14)
					{
						// BW 40, A band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x50);
					}
					else
					{
						// BW 40, G band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x40);
					}
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R69, 0x30);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R70, 0x30);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R73, 0x08);
				}
				else // BW 20
				{
					if (pAd->CommonCfg.Channel > 14)
					{
						// BW 20, A band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x60);
					}
					else
					{
						// BW 20, G band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x30);
					}
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R69, 0x28);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R70, 0x28);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R73, 0x08);
				}
				
			}
			else // RADAR_DETECTION
			{
				// save old value				
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R65, &pAd->CommonCfg._R65);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R66, &pAd->CommonCfg._R66);

				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->CommonCfg.R65);
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg.R66);
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x21);
			}		
			return;
}

#endif // #if defined(DFS_SUPPORT) || defined(CARRIER_DETECTION_SUPPORT)




